import logging, re
from functools import reduce

from peeringdb import resource, get_backend

try:
    input = raw_input
except NameError:
    pass


def split_ref(string):
    """ splits a string into (tag, id) """
    re_tag = re.compile('^(?P<tag>[a-zA-Z]+)[\s-]*(?P<pk>\d+)$')
    m = re_tag.search(string)
    if not m:
        raise ValueError("unable to split string '%s'" % (string, ))

    return (m.group('tag').lower(), int(m.group('pk')))


def pretty_speed(value):
    if not value:
        return ''
    try:
        value = int(value)
        if value >= 1000000:
            return "%dT" % (value / 10**6)
        elif value >= 1000:
            return "%dG" % (value / 10**3)
        else:
            return "%dM" % value
    except ValueError:
        return value


def prompt(msg, default=None):
    "Prompt for input"
    if default is not None:
        msg = '{} ({})'.format(msg, repr(default))
    msg = '{}: '.format(msg)
    try:
        s = input(msg)
    except KeyboardInterrupt:
        exit(1)
    except EOFError:
        s = ''
    if not s:
        s = default
    return s


class FieldGroups:
    "Partition a concrete's fields into groups based on type"
    GROUPS = ('scalars', 'one_refs', 'many_refs')

    def __init__(self, concrete):
        backend = get_backend()
        kinds = {kind: {} for kind in FieldGroups.GROUPS}
        fields = backend.get_fields(concrete)
        for field in fields:
            name = field.name
            related, multiple = backend.is_field_related(concrete, name)

            if related:
                if multiple:
                    group = kinds['many_refs']
                else:
                    group = kinds['one_refs']
            else:
                group = kinds['scalars']

            group[name] = field
        self._fields = kinds

        for kind, fs in kinds.items():
            setattr(self, kind, lambda _fs=fs: _fs.items())

    def __getitem__(self, k):
        return self._fields[k]

    def __contains__(self, k):
        return k in self._fields


def limit_mem(limit=(4 * 1024**3)):
    "Set soft memory limit"
    rsrc = resource.RLIMIT_DATA
    soft, hard = resource.getrlimit(rsrc)
    resource.setrlimit(rsrc, (limit, hard))  # 4GB
    softnew, _ = resource.getrlimit(rsrc)
    assert softnew == limit

    _log = logging.getLogger(__name__)
    _log.debug('Set soft memory limit: %s => %s', soft, softnew)



"""
Sync implementation module
"""
import logging
import datetime
from collections import defaultdict
from peeringdb.resource import get_resource
from peeringdb import get_backend
from peeringdb.util import FieldGroups


def _field_resource(B, concrete, field):
    return B.get_resource(B.get_field_concrete(concrete, field))


def clean_helper(B, obj, clean_func):
    """
    Clean object, intercepting and collecting any missing-relation or
    unique-constraint errors and returning the relevant resource ids/fields.

    Returns:
        - tuple: (<dict of non-unique fields>, <dict of missing refs>)
    """
    try:
        clean_func(obj)
    except B.validation_error() as e:
        # _debug.log_validation_errors(B, e, obj, k)

        # Check if it's a uniqueness or missing relation error
        fields = B.detect_uniqueness_error(e)
        missing = B.detect_missing_relations(obj, e)
        return fields, missing
    return (None, None)


def initialize_object(B, res, row):
    """
    Do a shallow initialization of an object

    Arguments:
        - row<dict>: dict of data like depth=1, i.e. many_refs are only ids
    """
    B = get_backend()
    field_groups = FieldGroups(B.get_concrete(res))

    try:
        obj = B.get_object(B.get_concrete(res), row['id'])
    except B.object_missing_error(B.get_concrete(res)):
        tbl = B.get_concrete(res)
        obj = tbl()

    # Set attributes, refs
    for fname, field in field_groups['scalars'].items():
        value = row.get(fname, getattr(obj, fname, None))
        value = B.convert_field(obj.__class__, fname, value)
        setattr(obj, fname, value)
    # _debug('res, row: %s, %s', res, row)

    # Already-fetched, and id-only refs
    fetched, dangling = defaultdict(dict), defaultdict(set)

    # To handle subrows that might be shallow (id) or deep (dict)
    def _handle_subrow(R, subrow):
        if isinstance(subrow, dict):
            pk = subrow['id']
            fetched[R][pk] = subrow
        else:
            pk = subrow
            dangling[R].add(pk)
        return pk

    for fname, field in field_groups['one_refs'].items():
        fieldres = _field_resource(B, B.get_concrete(res), fname)
        key = field.column
        subrow = row.get(key)
        if subrow is None:  # e.g. use "org" if "org_id" is missing
            key = fname
            subrow = row[key]
        pk = _handle_subrow(fieldres, subrow)
        setattr(obj, key, pk)

    for fname, field in field_groups['many_refs'].items():
        fieldres = _field_resource(B, B.get_concrete(res), fname)
        pks = [
            _handle_subrow(fieldres, subrow) for subrow in row.get(fname, [])
        ]

    return obj, fetched, dangling


def set_object_relations(B, res, obj, row):
    field_groups = FieldGroups(B.get_concrete(res))

    for fname, field in field_groups['many_refs'].items():
        fieldres = _field_resource(B, B.get_concrete(res), fname)
        pks = row.get(fname, [])
        objs = [B.get_object(B.get_concrete(fieldres), pk) for pk in pks]
        B.set_relation_many_to_many(obj, fname, objs)


def patch_object(B, res, obj, strip_tz):
    field_groups = FieldGroups(B.get_concrete(res))

    for fname in field_groups['scalars']:
        value = getattr(obj, fname)
        if strip_tz and isinstance(value, datetime.datetime):
            value = value.replace(tzinfo=None)
        B.update(obj, fname, value)




import yaml
from peeringdb import get_backend
from peeringdb.util import FieldGroups

# from peeringdb.debug import try_or_debug


# Wrap orm object into node for graph traversal
class YamlWrap(FieldGroups):
    def __init__(self, o, depth):
        self.tag = 'tag:yaml.org,2002:map'
        self.object = o
        self.depth = depth
        self.fields = FieldGroups(o.__class__)

    @staticmethod
    def _resolve_one(name, value, depth):
        if depth > 0:
            return YamlWrap(value, depth - 1)
        else:
            return value.id

    @staticmethod
    def _resolve_many(name, value, depth):
        if depth > 1:
            return [YamlWrap(o, depth - 1) for o in value.all()]
        elif depth == 1:
            return list(value.values_list('id', flat=True))

    def resolve(self, group, name, value):
        if group == 'scalars':
            return value
        elif group == 'one_refs':
            return YamlWrap._resolve_one(name, value, self.depth)
        elif group == 'many_refs':
            return YamlWrap._resolve_many(name, value, self.depth)
        else:
            raise ValueError(group)

    def field_values(self):
        for group in self.fields.GROUPS:
            if self.depth == 0 and group == 'many_refs':
                continue
            for name in self.fields[group]:
                value = self.resolve(group, name, getattr(self.object, name))
                yield name, value


def represent_wrapped(dumper, wrap):
    _dict = {}
    for name, value in wrap.field_values():
        _dict[name] = value
    alist = [(k, _dict[k]) for k in sorted(_dict)]
    return dumper.represent_mapping(wrap.tag, alist)


def default_representer(dumper, data):
    # Py2 workaround
    rep_func = getattr(dumper, 'represent_unicode', None)
    if rep_func is None:
        rep_func = dumper.represent_str
    return rep_func(str(data))


def _init():
    dumper = yaml.SafeDumper
    for cls in get_backend().CUSTOM_FIELDS:
        dumper.add_representer(cls, default_representer)
    dumper.add_representer(YamlWrap, represent_wrapped)


def dump(obj, depth, file):
    _init()
    if get_backend().is_concrete(type(obj)):
        obj = YamlWrap(obj, depth)
    yaml.safe_dump(obj, stream=file, default_flow_style=False)



"""
Module defining main interface classes for sync
"""
import logging
import threading
from itertools import chain
from datetime import datetime

from peeringdb import resource, get_backend
from peeringdb import _sync, _fetch, _config_logs

try:
    from peeringdb import _tasks_async as _tasks
except:
    from peeringdb import _tasks_sequential as _tasks


class _CancelSave(Exception):
    "Dummy exception to cancel transaction"
    pass


class Updater(object):
    """
    Class for updating resource data.

    Arguments:
        - fetcher: Fetcher instance for RPC calls
        - strip_tz<bool>: strip timezones from datetime values
        - dry_run<bool>: perform sync, but don't actually commit to
            database
    """

    def __init__(self, fetcher, strip_tz=True, dry_run=False, **_):
        self._fetcher = fetcher
        self.strip_tz = strip_tz
        self.dry_run = dry_run
        # allow dep. injection for testing
        self._ContextClass = UpdateContext
        # Must initialize logs now
        _config_logs()
        self._log = logging.getLogger(__name__)

    def update(self, res, pk, depth=1, since=None):
        """
        Try to sync an object to the local database, in case of failure
        where a referenced object is not found, attempt to fetch said
        object from the REST api
        """
        fetch = lambda: self._fetcher.fetch_latest(res, pk, 1, since=since)
        self._update(res, fetch, depth)

    def update_where(self, res, depth=0, since=None, **kwargs):
        "Like update() but uses WHERE-style args"
        fetch = lambda: self._fetcher.fetch_all_latest(res, 0, kwargs, since=since)
        self._update(res, fetch, depth)

    def update_all(self, rs=None, since=None):
        "Sync all objects for the relations rs (if None, sync all resources)"
        self._log.info("Updating resources: %s", ' '.join(r.tag for r in rs))

        if rs is None:
            rs = resource.all_resources()
        ctx = self._ContextClass(self)
        for r in rs:
            self._atomic_update(lambda: ctx.sync_resource(r, since=since))

    def _update(self, res, fetch_func, depth):
        ctx = self._ContextClass(self)
        data, e = fetch_func()
        if e: raise e
        self._log.info("Updates to be processed: {}".format(len(data)))
        self._atomic_update(lambda: ctx.sync_rows(res, data, depth + 1))

    def _atomic_update(self, sync_func):
        try:
            with get_backend().atomic_transaction():
                sync_func()
                if self.dry_run:
                    raise _CancelSave
                self._log.debug('Committing transaction')
        except _CancelSave:
            self._log.info('Transaction commit was cancelled (dry run)')
            pass


# Holds the state relevant to a single sync
# fetched: contains finished fetch-and-sync jobs
class UpdateContext(object):
    def __init__(self, updater):
        self.updater = updater
        self.fetcher = updater._fetcher
        self._log = updater._log
        self._jobs = {R: {}
                      for R in resource.all_resources()}, threading.Lock()
        self.disable_partial = True  # TODO
        self.start_time = datetime.now()

    @_tasks.run_task
    def sync_resource(self, res, since=None):
        self._log.info("Fetching & updating all: %s", res.tag)
        data, e = self.fetcher.fetch_all_latest(res, 0, since=since)
        self._log.info("Updates to be processed: %s", len(data))
        return _tasks.gather(self._schedule_rows(res, data, 0))


    # # Disabled, runaway memory
    # def sync_resources(self, rs):
    #     yield _tasks.gather(chain.from_iterable(self.sync_resource(res) for res in rs))

    @_tasks.run_task
    def sync_rows(self, res, rows, depth):
        return _tasks.gather(self._schedule_rows(res, rows, depth))

    def _schedule_rows(self, res, rows, depth):
        return [
            self.set_job((res, row['id']), self.sync_row,
                         (res, row, depth)) for row in rows
        ]

    def get_task(self, key):
        """Get a scheduled task, or none"""
        res, pk = key
        jobs, lock = self._jobs
        with lock:
            return jobs[res].get(pk)

    def set_job(self, key, func, args):
        """
        Get a scheduled task or set if none exists.

        Returns:
            - task coroutine/continuation
        """
        res, pk = key
        jobs, lock = self._jobs
        task = _tasks.UpdateTask(func(*args), key)
        with lock:
            job = jobs[res].get(pk)
            had = bool(job)
            if not job:
                job = task
                jobs[res][pk] = job
            else:
                task.cancel()
        self._log.debug('Scheduling: %s-%s (%s)', res.tag, pk,
                        'new task' if not had else 'dup')
        return job

    def pending_tasks(self, res):
        "Synchronized access to tasks"
        jobs, lock = self._jobs
        with lock:
            return jobs[res].copy()

    @_tasks.wrap_generator
    def fetch_and_index(self, fetch_func):
        "Fetch data with func, return dict indexed by ID"
        data, e = fetch_func()
        if e: raise e
        yield {row['id']: row for row in data}

    # Update an object after the fetch job finishes
    @_tasks.wrap_generator
    def update_after(self, res, pk, depth, fetch_job):
        data = yield fetch_job
        if data:
            row = data[pk]
        else:
            self._log.info('Fetched no data for %s-%s', res.tag, pk)
            data, e = self.fetcher.fetch_deleted(res, pk, 0)
            if not data:
                self._log.info(
                    'Fetched no deleted objects for %s-%s, aborting', res.tag,
                    pk)
                return
            row = data[0]
        yield self.sync_row(res, row, depth)

    @_tasks.wrap_generator
    def sync_row(self, res, row, depth):
        self._log.debug("sync_row(%s, %s, %s)", res, row['id'], depth)
        if self.disable_partial and depth > 0:
            raise ValueError('depth > 0 sync is disabled')

        B = get_backend()
        obj, fetched, dangling = _sync.initialize_object(B, res, row)

        # self._log.debug(' fetched: %s', fetched)
        # self._log.debug(' dangling: %s', dangling)

        def _have(R, pks):
            have = B.get_objects(B.get_concrete(R), pks)
            return set(have.values_list('id', flat=True))

        # Before attempting to set the related-object fields, ensure they are synced
        # Skip all ref'd objects that we already have, or have scheduled to update
        sync_jobs = []
        for R, sub in fetched.items():
            have = _have(R, set(sub.keys()))
            sync_jobs.extend(
                self.set_job((R, pk), self.sync_row, (R, subrow, depth - 1))
                for pk, subrow in sub.items() if not (pk in have))
        for R, pks in dangling.items():
            pks = pks.difference(_have(R, pks))
            pending = self.pending_tasks(R)
            needpks = []
            for pk in pks:
                if pk in pending:
                    sync_jobs.append(self.get_task((R, pk)))
                else:
                    needpks.append(pk)
            if not needpks:
                continue

            def fetch(_R=R, _pks=needpks):
                return self.fetcher.fetch_all_latest(_R, 0, dict(id__in=_pks))

            fetch_job = _tasks.UpdateTask(
                self.fetch_and_index(fetch), (R, None))

            sync_jobs.extend(
                self.set_job((R, pk), self.update_after,
                             (R, pk, depth - 1, fetch_job)) for pk in pks)

        # Detect integrity gaps
        dup_fields, missing = _sync.clean_helper(B, obj, B.clean)

        if dup_fields:
            # For non-uniques, try to update conflicting object
            field = dup_fields[0]
            value = getattr(obj, field)
            try:
                dup = B.get_object_by(B.get_concrete(res), field, value)
                dup_id = dup.id
                dup.delete(hard=True)
            except B.object_missing_error(B.get_concrete(res)):  # shouldn't happen
                raise Exception('internal error')

            self._log.debug('dup: %s = %s', field, value)

            def fetch():
                data, err = self.fetcher.fetch_latest(res, dup_id, 0)
                if isinstance(err, _fetch.NotFoundException):
                    # case where the local duplicate has been deleted
                    return {}, None
                return data, err

            fetch_job = _tasks.UpdateTask(
                self.fetch_and_index(fetch),
                (res, dup_id))
            job = self.set_job((res, dup_id), self.update_after,
                               (res, dup_id, 0, fetch_job))
            sync_jobs.append(job)

        if missing:
            # ignore expected missing refs
            for R, pks in missing.items():
                for pk in pks:
                    if pk not in dangling[R]:  # shouldn't happen
                        raise RuntimeError("Unexpected missing relation",
                                           (R, pk))

        # Preliminary save so this object exists for its dependencies
        _sync.patch_object(B, res, obj, self.updater.strip_tz)

        # Ignore new objects for consistency - TODO: test
        if obj.updated >= self.start_time:
            self._log.info('Ignoring object updated after sync began: (%s-%s)',
                           res.tag, row['id'])
            return

        if not self.updater.dry_run:
            B.save(obj)

        # Now, resolve refs and then save full object
        self._log.debug(' waiting for (%s-%s): %s', res.tag, row['id'],
                        sync_jobs)
        yield _tasks.gather(sync_jobs)
        self._log.debug("sync_row(%s, %s, %s) (resumed)", res.tag, row['id'],
                        depth)

        if self.updater.dry_run:
            return

        _sync.set_object_relations(B, res, obj, row)
        B.clean(obj)
        B.save(obj)



from __future__ import print_function
import os, sys
from argparse import ArgumentParser

import peeringdb
from peeringdb import config, commands

COMMANDS = {
    'get': commands.Get,
    'whois': commands.Whois,
    'sync': commands.Sync,
    'config': commands.CommandGroup({
        'set': commands.PromptConfig,
        'show': commands.DumpConfig,
        'list-codecs': commands.ListCodecs,
    }, help="Configuration management"),
    'drop-tables': commands.DropTables,
}


def check_load_config(config_dir):
    convert = False
    loaded = config.read_config(config_dir) or {}

    if config.detect_old(loaded):
        print(
            "Found config file with pre-0.7 schema; backing up and converting to new format"
        )
        convert = True
        cfg = config.convert_old(loaded)
    else:
        cfg = config.default_config()
        config.recursive_update(cfg, loaded)

    if convert:
        config.write_config(cfg, config_dir, backup_existing=True)
    return cfg


def main(args=sys.argv):
    peeringdb._config_logs()

    parser = ArgumentParser()
    parser.add_argument('--version', action='version',
                        version='%(prog)s ' + peeringdb.__version__)
    parser.add_argument(
        '-C', '--config-dir', default=os.environ.get(
            'PEERINGDB_HOME', config.DEFAULT_CONFIG_DIR),
        help="Directory containing configuration files")

    cmd = commands.CommandGroup(COMMANDS)
    cmd.add_arguments(parser)

    try:
        options = parser.parse_args(args[1:])
    except SystemExit as e:
        return e.code
    cfg = check_load_config(options.config_dir)

    handler = getattr(options, 'handler', options.usage_handler)
    try:
        return handler(config=cfg, **vars(options))
    except peeringdb._fetch.CompatibilityError as e:
        print(e, file=sys.stderr)
        return 1



"""
PeeringDB resource definitions
"""
from collections import OrderedDict

# Generate classes
_NAMES = OrderedDict([
    ('org', 'Organization'),
    ('fac', 'Facility'),
    ('net', 'Network'),
    ('ix', 'InternetExchange'),
    ('ixfac', 'InternetExchangeFacility'),
    ('ixlan', 'InternetExchangeLan'),
    ('ixpfx', 'InternetExchangeLanPrefix'),
    ('netfac', 'NetworkFacility'),
    ('netixlan', 'NetworkIXLan'),
    ('poc', 'NetworkContact'),
])

RESOURCES_BY_TAG = OrderedDict()

for tag, name in _NAMES.items():

    class Meta(type):
        def __repr__(cls, _name=name):
            return _name

    Class = Meta(name, (), {'tag': tag})

    RESOURCES_BY_TAG[tag] = Class
    locals()[name] = Class

is_resource_tag = RESOURCES_BY_TAG.__contains__


def get_resource(tag):
    return RESOURCES_BY_TAG[tag]


get = get_resource


def all_resources():
    return list(RESOURCES_BY_TAG.values())



from __future__ import print_function
import sys
import calendar
import logging
from argparse import ArgumentParser

import munge
import peeringdb
from peeringdb import resource, util, config as cfg
from peeringdb.client import Client
from peeringdb.whois import WhoisFormat
from peeringdb.output._yaml import dump
from peeringdb.config import prompt_config, write_config


def _handler(func):
    "Decorate a command handler"

    def _wrapped(*a, **k):
        r = func(*a, **k)
        if r is None: r = 0
        return r

    return staticmethod(_wrapped)


def add_subcommands(parser, commands):
    "Add commands to a parser"
    subps = parser.add_subparsers()
    for cmd, cls in commands:
        subp = subps.add_parser(cmd, help=cls.__doc__)
        add_args = getattr(cls, 'add_arguments', None)
        if add_args:
            add_args(subp)
        handler = getattr(cls, 'handle', None)
        if handler:
            subp.set_defaults(handler=handler)


class CommandGroup:
    """A group of nested subcommands."""

    def __init__(self, commands, help=None):
        """
        Arguments:
            - commands<dict>: dict of command names to handler classes
        """
        self.commands = commands
        if help:
            self.__doc__ = help

    def add_arguments(self, parser):
        add_subcommands(parser, self.commands.items())

        def _usage(**_):
            parser.print_usage()
            return 1

        parser.set_defaults(usage_handler=_usage)


class Get:
    "Get a resource"

    @staticmethod
    def add_arguments(parser):
        parser.add_argument('poids', nargs='+', help='Object IDs')
        parser.add_argument('--depth', '-D', default=0, type=int,
                            help='How many levels of nested objects to fetch')
        parser.add_argument('--output-format', '-O', default='yaml',
                            help='Output data format')
        parser.add_argument(
            '--remote', '-R', action='store_true', default=False,
            help='Fall back to API request if object is not found')

    @_handler
    def handle(config, poids, output_format, depth, remote, **_):
        client = Client(config)
        for poid in poids:
            (tag, pk) = util.split_ref(poid)
            res = resource.get_resource(tag)
            B = peeringdb.get_backend()
            try:
                obj = client.get(res, pk)
            except B.object_missing_error(B.get_concrete(res)) as e:
                if remote:
                    obj = client.fetch(res, pk, depth)[0]
                else:
                    print('Not found: {}-{}'.format(tag, pk), file=sys.stderr)
                    return 1

            dump(obj, depth, sys.stdout)


def _lookup_tag(tag, key, getfunc):
    if tag == 'as':
        return getfunc(resource.Network, asn=key)
    elif tag == 'ixnets':
        return getfunc(resource.Network, ix_id__in=key)
    else:
        return getfunc(resource.get_resource(tag), id=key)


class Whois:
    """Simulate a whois lookup;
       supports
            as<ASN> : query by AS;
            ixnets<net ID> : query networks on an IX
    """

    @staticmethod
    def add_arguments(parser):
        parser.add_argument(
            'poids', nargs='+',
            help='Object IDs, or { as<ASN> | ixnets<net ID> }...')

    @_handler
    def handle(config, poids, **_):
        client = Client(config)

        def get(res, **kwargs):
            objs = client.fetch_all(res, depth=2, **kwargs)
            if not objs:
                raise peeringdb.sync.NotFoundException
            return objs

        fmt = WhoisFormat()
        for poid in poids:
            (tag, key) = util.split_ref(poid)
            try:
                data = _lookup_tag(tag, key, get)
            except peeringdb.sync.NotFoundException:
                print('Not found: resource for {}={}'.format(tag, key),
                      file=sys.stderr)
                return 1
            fmt.display(tag, data[0])


class DumpConfig:
    """ Output current config """

    @staticmethod
    def add_arguments(parser):
        parser.add_argument('--output-format', default='yaml',
                            help='output data format')

    @_handler
    def handle(config, output_format, **_):
        codec = munge.get_codec(output_format)()
        codec.dump(config, sys.stdout)


class PromptConfig:
    """ Prompt for configuration values and save to a file """

    @staticmethod
    def add_arguments(parser):
        parser.add_argument('--output-format', default='yaml',
                            help='output data format')
        parser.add_argument('--defaults', '-n', action='store_true',
                            default=False,
                            help='do not prompt, just write defaults to file')

    @_handler
    def handle(config, defaults, config_dir, output_format, **_):
        if defaults:
            newconfig = config
            outdir = config_dir
        else:
            newconfig = cfg.prompt_config(cfg.CLIENT_SCHEMA, defaults=config)
            outdir = util.prompt('Output directory', config_dir)
        cfg.write_config(newconfig, outdir, codec=output_format)


class ListCodecs:
    "List available codecs"

    @_handler
    def handle(**_):
        print(' '.join(munge.codec.list_codecs()))


class Sync:
    "Synchronize local tables to PeeringDB"

    @staticmethod
    def add_arguments(parser):
        parser.add_argument('-v', '--verbose', action='count',
                            help='Be more verbose')
        parser.add_argument('-q', '--quiet', action='count',
                            help='Be more quiet')
        parser.add_argument('--dry-run', action='store_true', default=False,
                            help='Do not actually perform updates')
        # parser.add_argument('--only', action='append', default=[],
        #                     help='Only process this table')
        # parser.add_argument('--limit', type=int, default=0,
        #                     help="Limit objects retrieved, retrieve all if 0 (default)")
        parser.add_argument('--init', action='store_true', default=False,
                            help='Only initialize the database; do not sync')

    @_handler
    def handle(config, verbose, quiet, init, **kwargs):
        rs = resource.all_resources()
        # if only: rs = [resource.get_resource(tag) for tag in only]

        loglvl = 1 + (verbose or 0) - (quiet or 0)
        if loglvl > 1:
            peeringdb._config_logs(logging.DEBUG)
        if loglvl < 1:
            peeringdb._config_logs(logging.WARNING)

        client = Client(config, **kwargs)
        # todo verify table schema
        if init: return

        if loglvl >= 0:
            print("Syncing to", config['sync']['url'])

        client.update_all(rs)


class DropTables:
    "Drop all database tables"

    @_handler
    def handle(config, **_):
        client = Client(config)
        B = peeringdb.get_backend()
        B.delete_all()
